import re
import random
import asyncio
from pathlib import Path
from typing import Union, Optional
from io import BytesIO, TextIOWrapper

from nonebot.log import logger

from .schema import Badge
from .network import auto_retry
from .api import bg_url, safe_async_get
from .network.first_response import get_first_response

osufile = Path(__file__).parent / "osufile"
map_path = Path() / "data" / "osu" / "map"
user_cache_path = Path() / "data" / "osu" / "user"
badge_cache_path = Path() / "data" / "osu" / "badge"
team_cache_path = Path() / "data" / "osu" / "team"
api_ls = [
    "https://api.chimu.moe/v1/download/",
    "https://osu.direct/api/d/",
    "https://txy1.sayobot.cn/beatmaps/download/novideo/",
    "https://catboy.best/d/",
]
semaphore = asyncio.Semaphore(5)

if not map_path.exists():
    map_path.mkdir(parents=True, exist_ok=True)
if not user_cache_path.exists():
    user_cache_path.mkdir(parents=True, exist_ok=True)
if not badge_cache_path.exists():
    badge_cache_path.mkdir(parents=True, exist_ok=True)
if not team_cache_path.exists():
    team_cache_path.mkdir(parents=True, exist_ok=True)


async def download_map(setid: int) -> Optional[Path]:
    urls = [i + str(setid) for i in api_ls]
    logger.info(f"开始下载地图: <{setid}>")
    req = await get_first_response(urls)
    filename = f"{setid}.osz"
    filepath = map_path.parent / filename
    with open(filepath, "wb") as f:
        f.write(req.read())
    logger.info(f"地图: <{setid}> 下载完毕")
    return filepath


async def download_tmp_osu(map_id):
    url = f"https://osu.ppy.sh/osu/{map_id}"
    logger.info(f"开始下载谱面: <{map_id}>")
    req = await safe_async_get(url)
    filename = f"{map_id}.osu"
    filepath = map_path / filename
    chunk = req.read()
    with open(filepath, "wb") as f:
        f.write(chunk)
    return filepath


@auto_retry
async def download_osu(set_id, map_id):
    url = [
        f"https://osu.ppy.sh/osu/{map_id}",
        f"https://osu.direct/api/osu/{map_id}",
        f"https://catboy.best/osu/{map_id}",
    ]
    logger.info(f"开始下载谱面: <{map_id}>")
    async with semaphore:
        if req := await get_first_response(url):
            filename = f"{map_id}.osu"
            filepath = map_path / str(set_id) / filename
            with open(filepath, "wb") as f:
                f.write(req)
            return filepath
        else:
            raise Exception("下载出错，请稍后再试")


async def get_projectimg(url: str) -> BytesIO:
    if "avatar-guest.png" in url:
        url = "https://osu.ppy.sh/images/layout/avatar-guest.png"
    req = await safe_async_get(url)
    if not req or req.status_code >= 400:
        # todo 加个自创的错误图片
        req = await safe_async_get(random.choice(bg_url))
        if not req or req.status_code >= 400:
            raise Exception("图片下载失败")
    data = req.read()
    im = BytesIO(data)
    return im


async def get_pfm_img(url: str, cache_path: Path) -> BytesIO:
    cache_dir = cache_path.parent
    if not cache_dir.exists():
        cache_dir.mkdir(parents=True, exist_ok=True)
    if cache_path.exists():
        with cache_path.open("rb") as f:
            return BytesIO(f.read())
    async with asyncio.Semaphore(5):
        req = await safe_async_get(url)
    if not req or req.status_code >= 400:
        return BytesIO()
    image_data = req.content
    with cache_path.open("wb") as f:
        f.write(image_data)
    return BytesIO(image_data)


def re_map(file: Union[bytes, Path]) -> str:
    if isinstance(file, bytes):
        text = TextIOWrapper(BytesIO(file), "utf-8").read()
    else:
        with open(file, encoding="utf-8") as f:
            text = f.read()
    res = re.search(r"\d,\d,\"(.+)\"", text)
    bg = "mapbg.png" if not res else res.group(1).strip()
    if "/" in bg:
        bg = bg.split("/")[-1]
    return bg


async def make_badge_cache_file(badge: Badge):
    path = badge_cache_path / f"{hash(badge.description)}.png"
    badge_icon = await get_projectimg(badge.image_url)
    with open(path, "wb") as f:
        f.write(badge_icon.getvalue())


# 保存个人信息界面背景
async def save_info_pic(user: str, byt: bytes):
    path = user_cache_path / user
    if not path.exists():
        path.mkdir()
    with open(path / "info.png", "wb") as f:
        f.write(BytesIO(byt).getvalue())
